library(RGtk2)

.First.lib <- function(lib, pkg) {
       library.dynam("RImgViewer", pkg, lib)
}


gtkImageGetData <-
  function(widget,.flush=TRUE) {
    s<-.Call("S_gtk_image_get_array", widget,PACKAGE='RImgViewer')
    return(s)
    
  }

gtkImageSetData <-
  function(widget,ImData,.flush=TRUE) {
       s<-.Call("S_gtk_image_set_array", widget,ImData, PACKAGE='RImgViewer')
       return(s)     
  }


image.viewer <- function(imagefile=NULL){
  .markers.counter <- 0
  .Markers <- list()
  .Tags <- NULL
  .FileSel<-imagefile
  .CB<<-NULL
  unset.CB <- function(){
    sapply(.CB,function(x){
      e4$DisconnectCallback(x)
    })
    .CB<<-NULL
  }
  array.from.image <- function(only.visible=FALSE){
    z <- gtkImageGetData(l4)
    if (only.visible){
      s4$GetVadjustment()->va
      page <- gtkAdjustmentGetPageSize(va)
      start <-  gtkAdjustmentGetValue(va)
      top <- gtkAdjustmentGetUpper(va)
      bottom <- gtkAdjustmentGetLower(va)
      clamp.vrange <- as.integer(c(start/(top-bottom),(start+page)/(top-bottom))*dim(z)[2])
      s4$GetHadjustment()->ha
      page <- gtkAdjustmentGetPageSize(ha)
      start <-  gtkAdjustmentGetValue(ha)
      top <- gtkAdjustmentGetUpper(ha)
      bottom <- gtkAdjustmentGetLower(ha)
      clamp.hrange <- as.integer(c(start/(top-bottom),(start+page)/(top-bottom))*dim(z)[1])
      z <- z[clamp.hrange[1]:clamp.hrange[2],clamp.vrange[1]:clamp.vrange[2],]
    }
    return(z)
  }
  array.to.image <- function(data){
    .Tags<<-NULL
    .Markers<<-list()
    return(gtkImageSetData(l4,data))
  }
  
  start.stroke<-function(win, arg) {
    .Click.Inside.Area<<-TRUE
    xa<<-as.double(gdkEventButtonGetX(arg))
    ya<<-as.double(gdkEventButtonGetY(arg))
    xb<<-xa
    yb<<-ya
    return(TRUE)
  }
  
  line.end.stroke<-function(win, arg) {
    if(!.Click.Inside.Area) return(NULL)
    .Click.Inside.Area <<- FALSE
    unset.CB()
    xb<<-as.double(gdkEventButtonGetX(arg))
    yb<<-as.double(gdkEventButtonGetY(arg))
    gdkDrawLine(l4$GetWindow(),N1,as.integer(xa),as.integer(ya),as.integer(xb),as.integer(yb))
    add.marker("line",c(xa,ya,xb,yb))
    return(TRUE)
  }

  line.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      gdkDrawLine(l4$GetWindow(),N,as.integer(xa),as.integer(ya),as.integer(xb),as.integer(yb))
      xb<<-as.double(gdkEventMotionGetX(arg))
      yb<<-as.double(gdkEventMotionGetY(arg))
      gdkDrawLine(l4$GetWindow(),N,as.integer(xa),as.integer(ya),as.integer(xb),as.integer(yb))
    }
  }

  add.line.cb <- function(...){
    unset.CB()
    .CB<<-c(e4$AddCallback("motion_notify_event",line.stroke),e4$AddCallback("button-press-event",start.stroke),e4$AddCallback("button-release-event",line.end.stroke))
  }

  needle.end.stroke<-function(win, arg) {
    if(!.Click.Inside.Area) return(NULL)
    .Click.Inside.Area <<- FALSE
    unset.CB()
    s4[["hadjustment"]][["upper"]]->x1
    s4[["hadjustment"]][["lower"]]->x2
    s4[["vadjustment"]][["upper"]]->y1
    s4[["vadjustment"]][["lower"]]->y2
    xb<<-as.double(gdkEventButtonGetX(arg))
    yb<<-as.double(gdkEventButtonGetY(arg))
    XY <- .get.rot((yb-ya)/(xb-xa),x1,x2,y1,y2)
    gdkDrawLine(l4$GetWindow(),N1,as.integer(XY[1]),as.integer(XY[2]),as.integer(XY[3]),as.integer(XY[4]))
    gdkDrawLine(l4$GetWindow(),N1,as.integer(XY[5]),as.integer(XY[6]),as.integer(XY[7]),as.integer(XY[8]))
    add.marker("needle",c(xa,ya,XY[9]))
    return(TRUE)
  }

  .get.rot <- function(phi,x1,x2,y1,y2,x0=xa,y0=ya) {
    if (is.na(phi)) { phi <- 0 }
    if (abs(phi)>1) {
        X0 <- x0-y0/phi
        Y0 <- y2
        X1 <- x0 + (y1-y0)/phi
        Y1 <- y1
        Xa <- x1
        Ya <- y0-(x1-x0)/phi
        Xb <- x2
        Yb <- y0+x0/phi
      } else {
        X0 <- x1
        Y0 <- y0+(x1-x0)*phi
        X1 <- x2
        Y1 <- y0-x0*phi
        Xa <- x0+y0*phi
        Ya <- y2
        Xb <- x0 - (y1-y0)*phi
        Yb <- y1
      }
    return(c( X0,Y0,X1,Y1,Xa,Ya,Xb,Yb,phi))
  }
  
  needle.start.stroke<-function(win, arg) {
    .Click.Inside.Area<<-TRUE
    xa<<-as.double(gdkEventButtonGetX(arg))
    ya<<-as.double(gdkEventButtonGetY(arg))
    xb<<-xa
    yb<<-ya
    s4[["hadjustment"]][["upper"]]->x1
    s4[["hadjustment"]][["lower"]]->x2
    s4[["vadjustment"]][["upper"]]->y1
    s4[["vadjustment"]][["lower"]]->y2
    XY <- .get.rot((yb-ya)/(xb-xa),x1,x2,y1,y2)
    gdkDrawLine(l4$GetWindow(),N,as.integer(XY[1]),as.integer(XY[2]),as.integer(XY[3]),as.integer(XY[4]))
    gdkDrawLine(l4$GetWindow(),N,as.integer(XY[5]),as.integer(XY[6]),as.integer(XY[7]),as.integer(XY[8]))
    return(TRUE)
  }

  needle.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      s4[["hadjustment"]][["upper"]]->x1
      s4[["hadjustment"]][["lower"]]->x2
      s4[["vadjustment"]][["upper"]]->y1
      s4[["vadjustment"]][["lower"]]->y2
      XY <- .get.rot((yb-ya)/(xb-xa),x1,x2,y1,y2)
      gdkDrawLine(l4$GetWindow(),N,as.integer(XY[1]),as.integer(XY[2]),as.integer(XY[3]),as.integer(XY[4]))
      gdkDrawLine(l4$GetWindow(),N,as.integer(XY[5]),as.integer(XY[6]),as.integer(XY[7]),as.integer(XY[8]))
      xb<<-as.double(gdkEventMotionGetX(arg))
      yb<<-as.double(gdkEventMotionGetY(arg))
      XY <- .get.rot((yb-ya)/(xb-xa),x1,x2,y1,y2)
      gdkDrawLine(l4$GetWindow(),N,as.integer(XY[1]),as.integer(XY[2]),as.integer(XY[3]),as.integer(XY[4]))
      gdkDrawLine(l4$GetWindow(),N,as.integer(XY[5]),as.integer(XY[6]),as.integer(XY[7]),as.integer(XY[8]))
    }
    return(TRUE)
  }

  add.needle.cb <- function(...){
    unset.CB()
    .CB<<-c(e4$AddCallback("motion_notify_event",needle.stroke),e4$AddCallback("button-press-event",needle.start.stroke),e4$AddCallback("button-release-event",needle.end.stroke))
  }
  
  rect.end.stroke<-function(win, arg) {
    if(!.Click.Inside.Area) return(NULL)
    .Click.Inside.Area <<- FALSE
    unset.CB()  
    xb<<-as.double(gdkEventButtonGetX(arg))
    yb<<-as.double(gdkEventButtonGetY(arg))
    gdkDrawRectangle(l4$GetWindow(),N1,FALSE,as.integer(min(xa,xb)),as.integer(min(ya,yb)),as.integer(abs(xb-xa)),as.integer(abs(yb-ya)))
    add.marker("rect",c(xa,ya,xb,yb))
    return(TRUE)
  }
  
  rect.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      gdkDrawRectangle(l4$GetWindow(),N,FALSE,as.integer(min(xa,xb)),as.integer(min(ya,yb)),as.integer(abs(xb-xa)),as.integer(abs(yb-ya)))
      xb<<-as.double(gdkEventMotionGetX(arg))
      yb<<-as.double(gdkEventMotionGetY(arg))
      gdkDrawRectangle(l4$GetWindow(),N,FALSE,as.integer(min(xa,xb)),as.integer(min(ya,yb)),as.integer(abs(xb-xa)),as.integer(abs(yb-ya)))
    }
    return(TRUE)
  }
  
  add.rect.cb <- function(...){
    unset.CB()
    .CB<<-c(e4$AddCallback("motion_notify_event",rect.stroke),e4$AddCallback("button-press-event",start.stroke),e4$AddCallback("button-release-event",rect.end.stroke))
  }
  
  circle.end.stroke<-function(win, arg) {
    if(!.Click.Inside.Area) return(NULL)
    .Click.Inside.Area <<- FALSE
    unset.CB() 
    xb<<-as.double(gdkEventButtonGetX(arg))
    yb<<-as.double(gdkEventButtonGetY(arg))
    gdkDrawArc(l4$GetWindow(),N1,FALSE,as.integer(xa-abs(xa-xb)),as.integer(ya-abs(ya-yb)),as.integer(2*abs(xa-xb)),as.integer(2*abs(ya-yb)),as.integer(0),as.integer(23040))  
    add.marker("circle",c(xa,ya,abs(xa-xb),abs(ya-yb)))
    return(TRUE)
  }
  circle.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      gdkDrawArc(l4$GetWindow(),N,FALSE,as.integer(xa-abs(xa-xb)),as.integer(ya-abs(ya-yb)),as.integer(2*abs(xa-xb)),as.integer(2*abs(ya-yb)),as.integer(0),as.integer(23040))      
      xb<<-as.double(gdkEventMotionGetX(arg))
      yb<<-as.double(gdkEventMotionGetY(arg))
      gdkDrawArc(l4$GetWindow(),N,FALSE,as.integer(xa-abs(xa-xb)),as.integer(ya-abs(ya-yb)),as.integer(2*abs(xa-xb)),as.integer(2*abs(ya-yb)),as.integer(0),as.integer(23040))  
      
    }
    return(TRUE)
  }
  add.circle.cb <- function(...){
    unset.CB()
    .CB<<-c(e4$AddCallback("motion_notify_event",circle.stroke),e4$AddCallback("button-press-event",start.stroke),e4$AddCallback("button-release-event",circle.end.stroke))
  }
  
  polygon.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      k1 <- .Markers[[length(.Markers)]][[2]]
#      z1<-list(k1[,1],c(xa,ya),k1[,dim(k1)[2]])
#      gdkDrawPolygon(l4$GetWindow(),N,FALSE,z1)
      xa<<-as.double(gdkEventMotionGetX(arg))
      ya<<-as.double(gdkEventMotionGetY(arg))
#      z1<-list(k1[,1],c(xa,ya),k1[,dim(k1)[2]])
#      gdkDrawPolygon(l4$GetWindow(),N,FALSE,z1)
    }
    return(TRUE)
  }
  
  polygon.reshape <- function(){
      if (!is.na(match(".TEMPORARY.polygon",.Tags))){
        z <- .Markers[[length(.Markers)]][[2]]
        l<-dim(z)
        if (l[2]==1) {
          drop.marker(".TEMPORARY.polygon")
          unset.CB()
          return(NULL)
        }
        .Markers[[length(.Markers)]][[2]]<<-z[,-l[2],drop=F]
        l4$Unmap()
        l4$Map()
      }
      return(TRUE)
    }

  polygon.complete <- function(){
      if (!is.na(match(".TEMPORARY.polygon",.Tags))){     
        .Tags[length(.Tags)]<<-paste("polygon",.markers.counter,sep="")
        .markers.counter<<-.markers.counter+1
      }
      unset.CB()
  }

  polygon.key <- function(win,arg){
    z <- gdkEventKeyGetKeyval(arg)
    if (z==65293) {
                                        #Enter - polygon selection complete
      polygon.complete()
    }
    if (z==65307){
                                         #Esc - last selected vertex discarded
      polygon.reshape()
    }                  
  }
  
  polygon.add.vertex <- function(win,arg){
    xa<<-as.double(gdkEventButtonGetX(arg))
    ya<<-as.double(gdkEventButtonGetY(arg))
    ba<-gdkEventButtonGetButton(arg)
    .Click.Inside.Area<<-FALSE
    switch(ba,
           {if (is.na(match(".TEMPORARY.polygon",.Tags))){
             add.marker("polygon", matrix(c(xa,ya),nrow=2),".TEMPORARY.polygon")
           } else {
             z <- .Markers[[length(.Markers)]][[2]]
             .Markers[[length(.Markers)]][[2]]<<-cbind(z,c(xa,ya))
           }
            l4$Unmap()
            l4$Map()
            e4["has-focus"]<-TRUE
          },
           {
             polygon.complete()
              },
           {
             polygon.reshape()
           }
           )

  }
  
  add.polygon.cb <- function(...){
    unset.CB()
    .Click.Inside.Area<<-FALSE
    .CB<<-c(e4$AddCallback("motion_notify_event",polygon.stroke),e4$AddCallback("button-press-event",start.stroke),e4$AddCallback("button-release-event",polygon.add.vertex),e4$AddCallback("key-press-event",polygon.key))
  }

  points.stroke <- function(win,arg){
    if (.Click.Inside.Area){
      k1 <- .Markers[[length(.Markers)]][[2]]
#      z1<-list(k1[,1],c(xa,ya),k1[,dim(k1)[2]])
#      gdkDrawPolygon(l4$GetWindow(),N,FALSE,z1)
      xa<<-as.double(gdkEventMotionGetX(arg))
      ya<<-as.double(gdkEventMotionGetY(arg))
#      z1<-list(k1[,1],c(xa,ya),k1[,dim(k1)[2]])
#      gdkDrawPolygon(l4$GetWindow(),N,FALSE,z1)
    }
    return(TRUE)
  }
  
  points.reshape <- function(){
      if (!is.na(match(".TEMPORARY.points",.Tags))){
        z <- .Markers[[length(.Markers)]][[2]]
        l<-dim(z)
        if (l[2]==1) {
          drop.marker(".TEMPORARY.points")
          unset.CB()
          return(NULL)
        }
        .Markers[[length(.Markers)]][[2]]<<-z[,-l[2],drop=F]
        l4$Unmap()
        l4$Map()
      }
      return(TRUE)
    }

  points.complete <- function(){
      if (!is.na(match(".TEMPORARY.points",.Tags))){     
        .Tags[length(.Tags)]<<-paste("points",.markers.counter,sep="")
        .markers.counter<<-.markers.counter+1
      }
      unset.CB()
  }

  points.key <- function(win,arg){
    z <- gdkEventKeyGetKeyval(arg)
    if (z==65293) {
                                        #Enter - points selection complete
      points.complete()
    }
    if (z==65307){
                                         #Esc - last selected vertex discarded
      points.reshape()
    }                  
  }
  
  points.add.vertex <- function(win,arg){
    xa<<-as.double(gdkEventButtonGetX(arg))
    ya<<-as.double(gdkEventButtonGetY(arg))
    ba<-gdkEventButtonGetButton(arg)
    .Click.Inside.Area<<-FALSE
    switch(ba,
           {if (is.na(match(".TEMPORARY.points",.Tags))){
             add.marker("points", matrix(c(xa,ya),nrow=2),".TEMPORARY.points")
           } else {
             z <- .Markers[[length(.Markers)]][[2]]
             .Markers[[length(.Markers)]][[2]]<<-cbind(z,c(xa,ya))
           }
            l4$Unmap()
            l4$Map()
            e4["has-focus"]<-TRUE
          },
           {
             points.complete()
              },
           {
             points.reshape()
           }
           )

  }
  
  
  add.points.cb <- function(...){
    unset.CB()
    .Click.Inside.Area<<-FALSE
    .CB<<-c(e4$AddCallback("motion_notify_event",points.stroke),e4$AddCallback("button-press-event",start.stroke),e4$AddCallback("button-release-event",points.add.vertex),e4$AddCallback("key-press-event",points.key))
  }

  drop.marker <- function(tag){
    z <- match(tag,.Tags)
    if(!is.na(z)){
      .Tags<<-.Tags[-1*z]
      .Markers[[z]]<<-NULL
      l4$Unmap()
      l4$Map()
    }
  }
  
  add.marker <- function(type,data,tag=NULL){
    .markers.counter<<-.markers.counter+1
    if (is.null(tag)){
      tag <- paste(type,.markers.counter,sep="")
    } else {
      drop.marker(tag)
    }
    .Markers<<-append(.Markers,list(list(type,data)))
    .Tags<<-append(.Tags,tag)
  }
  
  get.marker.data <- function(tag){
     z <- match(tag,.Tags)
     return(.Markers[[z]][[2]])
  }
  
  list.markers <- function(){
    return(.Tags)
  }
  
  list.markers.cb <- function(...){
    drop.mark.sub$GetChildren()->l
    sapply(l,function(x){x$Destroy()})
    sapply(.Tags,function(x){
      drop.mark.sub$Append(z <- gtkMenuItem(x))
      z$AddCallback("activate",function(f,...){
        .CL<-f$GetChildren()[[1]]$GetText()
        drop.marker(.CL)
      })
    })
  }
  FO.ok.button.clicked <- function(...){
    .FileSel <<- F1.w$GetFilename()
    F1.w$Destroy()    
    l4$Destroy()
    .Tags<<-NULL
    .Markers<<-list()
    CleanMe <- NULL
#check for compressed images
    if (regexpr(".*\\.bz2$",.FileSel)>0) {
      if (system(paste("bzcat",.FileSel,">",CleanMe <- tempfile()),intern=F)>0) {
        stop("Not a compressed image")
      }
      .FileSel <- CleanMe
    }
    if (regexpr(".*\\.gz$",.FileSel)>0) {
      if (system(paste("zcat",.FileSel,">",CleanMe <- tempfile()),intern=F)>0) {
        stop("Not a compressed image")
      }
      .FileSel <- CleanMe
    }    
    l4<<-gtkImageNewFromFile(.FileSel)
    if (!is.null(CleanMe)) {
      unlink(CleanMe)
    }
    l4$SetEvents(c("exposure-mask","button1-motion-mask","button-press-mask","button-release-mask","key-press-mask","enter-notify-mask","leave-notify-mask","focus-change-mask"))
    l4$SetAlignment(0,0)
    e4$Add(l4)
    l4$AddCallback("expose-event",expose.cb)
    e4$ShowAll()
  }

  FO.cancel.button.clicked <- function(...) {
    F1.w$Destroy()
  }
  open.file.cb <- function(...){
    F1.w <<- gtkFileSelection("Open Image File")
    F1.w$GetOkButton()$AddCallback("clicked",FO.ok.button.clicked)
    F1.w$GetCancelButton()$AddCallback("clicked",FO.cancel.button.clicked)
    F1.w$Show()
  }
  quit.cb <- function(...){
    w4$Destroy()
  }
  
  expose.cb <- function(...){
    gdkFlush()
    gtkAddIdle(function(...){
      lapply(.Markers,function(x){
        switch(x[[1]],
               "line"={
                 gdkDrawLine(l4$GetWindow(),N1,as.integer(x[[2]][1]),as.integer(x[[2]][2]),as.integer(x[[2]][3]),as.integer(x[[2]][4]))                 
               },
               "circle"={
                 gdkDrawArc(l4$GetWindow(),N1,FALSE,as.integer(x[[2]][1]-x[[2]][3]),as.integer(x[[2]][2]-x[[2]][4]),as.integer(2*x[[2]][3]),as.integer(2*x[[2]][4]),as.integer(0),as.integer(23040))
               },
               "rect"={
                 gdkDrawRectangle(l4$GetWindow(),N1,FALSE,min(x[[2]][1],x[[2]][3]),min(x[[2]][2],x[[2]][4]),abs(x[[2]][3]-x[[2]][1]),abs(x[[2]][4]-x[[2]][2]))
               },
               "polygon"={
                 gdkDrawPolygon(l4$GetWindow(),N1,FALSE,apply(x[[2]],2,function(y){return(as.list(y))}))
               },
               "points"={
                 apply(x[[2]],2,function(y){
                   gdkDrawArc(l4$GetWindow(),N1,TRUE,as.integer(y[1]),as.integer(y[2]),5,5,as.integer(0),as.integer(23040))})
                 },
               "needle"={
                 s4[["hadjustment"]][["upper"]]->x1
                 s4[["hadjustment"]][["lower"]]->x2
                 s4[["vadjustment"]][["upper"]]->y1
                 s4[["vadjustment"]][["lower"]]->y2
                 XY <- .get.rot(x[[2]][3],x1,x2,y1,y2,x0=x[[2]][1],y0=x[[2]][2])
                 gdkDrawLine(l4$GetWindow(),N1,as.integer(XY[1]),as.integer(XY[2]),as.integer(XY[3]),as.integer(XY[4]))
                 gdkDrawLine(l4$GetWindow(),N1,as.integer(XY[5]),as.integer(XY[6]),as.integer(XY[7]),as.integer(XY[8]))
                 
               })
      })
      return(FALSE)
    })
  }
  w4<-gtkWindow()
  if (is.null(.FileSel)) {
    gtkImageNew()->l4
  } else {
    gtkImageNewFromFile(.FileSel)->l4
  }
  l4["xalign"] <- 0
  l4["yalign"] <- 0
  w4$Add(vbox <- gtkVBox())
  vbox$SetHomogeneous(FALSE)
  vbox$Add(menubar <- gtkMenuBar())

  vbox$SetChildPacking(menubar,FALSE,FALSE,0,as.integer(0))
  vbox$Add(s4<-gtkScrolledWindow())
  menubar$Add(file1<-gtkMenuItem("File"))
  file1$SetSubmenu(file.sub <- gtkMenu())
  file.sub$Append(open.w <- gtkMenuItem("Open Image..."))
  open.w$AddCallback("activate",open.file.cb)
  file.sub$Append(quit.w <- gtkMenuItem("Quit"))
  quit.w$AddCallback("activate",quit.cb)
  menubar$Add(marker1<-gtkMenuItem("Markers"))
  marker1$SetSubmenu(marker.sub <- gtkMenu())
#  marker.sub$Append(drop.mark.w <- gtkMenuItem("Select Marker Color"))
#  marker.sub$Append(sep <- gtkMenuItem())
  marker.sub$Append(line.w <- gtkMenuItem("Add Line"))
  line.w$AddCallback("activate",add.line.cb)
  marker.sub$Append(rect.w <- gtkMenuItem("Add Rectangle"))
  rect.w$AddCallback("activate",add.rect.cb)
  marker.sub$Append(polyl.w <- gtkMenuItem("Add Polygon"))
  polyl.w$AddCallback("activate",add.polygon.cb)
  marker.sub$Append(points.w <- gtkMenuItem("Add Points"))
  points.w$AddCallback("activate",add.points.cb)
  marker.sub$Append(circle.w <- gtkMenuItem("Add Circle"))
  circle.w$AddCallback("activate",add.circle.cb)
  marker.sub$Append(needle.w <- gtkMenuItem("Add Needle"))
  needle.w$AddCallback("activate",add.needle.cb) 
  marker.sub$Append(sep <- gtkMenuItem())
  marker.sub$Append(drop.mark.w <- gtkMenuItem("Drop Markers..."))
  drop.mark.w$AddCallback("activate",list.markers.cb)
  drop.mark.w$SetSubmenu(drop.mark.sub <- gtkMenu())

  gtkEventBox()->e4
  l4$SetEvents(c("exposure-mask","button1-motion-mask","button-press-mask","button-release-mask","key-press-mask","enter-notify-mask","leave-notify-mask","focus-change-mask"))
  e4$SetEvents(c("exposure-mask","button1-motion-mask","button-press-mask","button-release-mask","key-press-mask","enter-notify-mask","leave-notify-mask","focus-change-mask"))
  e4["can-focus"]<-TRUE
  e4["has-focus"]<-TRUE
  s4$AddWithViewport(e4)
  e4$Add(l4)
  l4$AddCallback("expose-event",expose.cb)
  w4$ShowAll()
  e4["has-focus"]<-TRUE
  N<-gdkGC(l4$GetWindow())
  N$SetFunction(GdkFunction["xor"])
  N$SetRgbFgColor(c(0,65000,0))
  N$SetLineAttributes(1,GdkLineStyle["solid"],GdkCapStyle["butt"],GdkJoinStyle["miter"])
  N$SetFill(GdkFill["solid"])
  N1<-gdkGC(l4$GetWindow())
  N1$SetFunction(GdkFunction["copy"])
  N1$SetRgbFgColor(c(0,65000,0))
  N1$SetLineAttributes(1,GdkLineStyle["solid"],GdkCapStyle["butt"],GdkJoinStyle["miter"])
  N1$SetFill(GdkFill["solid"])
  invisible(list(array.to.image=array.to.image,array.from.image=array.from.image,add.marker=add.marker,drop.marker=drop.marker,get.marker.data=get.marker.data,list.markers=list.markers,debug=environment()))
  
}
#library(RImgViewer)
#image.viewer("/home/mike//OriondeepfieldM_gendler_big.jpg")->jj
